/*
 * Copyright (C) 2024 iopsys Software Solutions AB
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2.1
 * as published by the Free Software Foundation
 *
 *      Author: Suvendhu Hansa <suvendhu.hansa@iopsys.eu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/uloop.h>
#include <libbbfdm-ubus/bbfdm-ubus.h>

#include "users.h"
#include "helper.h"

extern DM_MAP_OBJ tDynamicObj[];
static struct bbfdm_context bbfdm_ctx = {0};
static struct ubus_event_handler ev = {0};

static void get_user_shell_name(const char *username, char **shell)
{
	char cmd[128] = {0};

	if (username == NULL || shell == NULL)
		return;

	snprintf(cmd, sizeof(cmd), "cat /etc/passwd | grep -E \"^%s:\" | cut -d: -f7", username);
	FILE *pp = popen(cmd, "r"); // flawfinder: ignore
	if (pp != NULL) {
		char line[2056] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			if (DM_STRCMP(line, "/bin/false") != 0) {
				char *pos = strrchr(line, '/');
				*shell = pos ? strdup(pos + 1) : strdup(line);
			}
		}
		pclose(pp);
	}
}

static void get_perticipated_groups(const char *username, char **groups)
{
	char cmd[128] = {0};

	if (username == NULL || groups == NULL)
		return;

	snprintf(cmd, sizeof(cmd), "groups %s", username);
	FILE *pp = popen(cmd, "r"); // flawfinder: ignore
	if (pp != NULL) {
		char line[2056] = {0};

		if (fgets(line, sizeof(line), pp) != NULL) {
			remove_new_line(line);
			if (DM_STRNCMP(line, "groups:", 7) != 0) {
				*groups = strdup(line);
			}
		}
		pclose(pp);
	}
}

static void find_unique_role_id(char **id)
{
	unsigned int i;

	if (id == NULL)
		return;

	*id = NULL;

	for (i = 0; i <= 0xFFFFFFFFU; i++) {
		char str_role[64] = {0};

		snprintf(str_role, sizeof(str_role), "%u", i);

		if (check_section_exist_by_option(STD_UCI_PATH, "users", "role", "role_id", str_role) == false &&
		    check_section_exist_by_option(DMMAP_UCI_PATH, DMMAP_SYSTEM_USER_FILE, "role", "role_id", str_role) == false) {
			*id = strdup(str_role);
			break;
		}
	}

	return;
}

static void sync_system_users_groups(FILE *pp, const char *package, const char *dmmap_package, const char *section_type, bool user_cfg)
{
	char line[65] = {0};

	if (pp == NULL || package == NULL || dmmap_package == NULL || section_type == NULL)
		return;

	while (fgets(line, sizeof(line), pp) != NULL) {
		struct uci_section *dmmap_sect = NULL;
		char name[70] = {0};

		remove_new_line(line);
		if (!user_cfg) {
			snprintf(name, sizeof(name), "%s%s", GROUP_PREFIX, line);
		} else {
			snprintf(name, sizeof(name), "%s", line);
		}

		replace_special_char(name, '_');
		if (get_section_from_config(STD_UCI_PATH, package, section_type, name) != NULL) {
			continue;
		}

		if (check_section_exist_by_option(DMMAP_UCI_PATH, dmmap_package, section_type, "section_name", name) == false) {
			// add to dmmap if not present in users as well as dmmap_system_users file
			usermgr_uci_init(DMMAP_UCI_PATH);

			dmmap_sect = usermgr_uci_add_section(dmmap_package, section_type);
			if (dmmap_sect == NULL) {
				usermgr_uci_fini(NULL);
				continue;
			}

			usermgr_uci_set_value_by_section(dmmap_sect, "section_name", name);
			usermgr_uci_set_value_by_section(dmmap_sect, "static", "1");

			if (user_cfg) {
				char *shell = NULL, *group = NULL;
				get_user_shell_name(name, &shell);
				if (shell) {
					usermgr_uci_set_value_by_section(dmmap_sect, "shell", shell);
					free(shell);
				}

				get_perticipated_groups(line, &group);
				if (group) {
					char *pch = NULL, *spch = NULL, lbuf[1024] = {0};
					unsigned pos = 0;
					for (pch = strtok_r(group, " ", &spch); pch != NULL; pch = strtok_r(NULL, " ", &spch)) {
						if (strlen(pch) < (1024 - strlen(lbuf) - strlen(GROUP_PREFIX))) {
							replace_special_char(pch, '_');
							pos += snprintf(&lbuf[pos], sizeof(lbuf) - pos, "%s%s,", GROUP_PREFIX, pch);
						}
					}

					if (pos)
						lbuf[pos - 1] = 0;

					usermgr_uci_set_value_by_section(dmmap_sect, "member_groups", lbuf);
					free(group);
				}
			}

			usermgr_uci_fini(dmmap_package);
		}
	}
}

static void sync_system_user(void)
{
	char cmd[256] = {0};

	snprintf(cmd, sizeof(cmd), "cat /etc/passwd | cut -d: -f1");

	FILE *pp = popen(cmd, "r"); // flawfinder: ignore
	if (pp != NULL) {
		sync_system_users_groups(pp, "users", DMMAP_SYSTEM_USER_FILE, "user", true);
		pclose(pp);
	}
}

static void sync_system_group(void)
{
	char cmd[256] = {0};

	snprintf(cmd, sizeof(cmd), "cat /etc/group | cut -d: -f1");

	FILE *pp = popen(cmd, "r"); // flawfinder: ignore
	if (pp != NULL) {
		sync_system_users_groups(pp, "users", DMMAP_SYSTEM_USER_FILE, "group", false);
		pclose(pp);
	}
}

static void sync_static_role(void)
{
	DIR *directory;

	directory = opendir("/etc/users/roles");
	if (directory == NULL) {
		return;
	}

	struct dirent *file;
	char role_file[256] = {0};

	while ((file = readdir(directory)) != NULL) {
		char *tmp = NULL;

		snprintf(role_file, sizeof(role_file), "%s", file->d_name);
		tmp = strrchr(role_file, '.');
		if (DM_STRCMP(tmp, ".json") != 0) {
			continue;
		}

		*tmp = '\0';
		if (check_section_exist_by_option(STD_UCI_PATH, "users", "role", "name", role_file) == true) {
			continue;
		}

		if (check_section_exist_by_option(DMMAP_UCI_PATH, DMMAP_SYSTEM_USER_FILE, "role", "name", role_file) == false) {
			// add to dmmap if not present in users as well as dmmap_system_users file
			struct uci_section *dmmap_sect = NULL;
			char *role_id = NULL;

			find_unique_role_id(&role_id);
			if (role_id == NULL)
				continue;

			usermgr_uci_init(DMMAP_UCI_PATH);
			dmmap_sect = usermgr_uci_add_section(DMMAP_SYSTEM_USER_FILE, "role");
			if (dmmap_sect == NULL) {
				free(role_id);
				usermgr_uci_fini(NULL);
				continue;
			}

			usermgr_uci_set_value_by_section(dmmap_sect, "enabled", "1");
			usermgr_uci_set_value_by_section(dmmap_sect, "role_id", role_id);
			usermgr_uci_set_value_by_section(dmmap_sect, "static", "1");
			usermgr_uci_set_value_by_section(dmmap_sect, "name", role_file);

			free(role_id);
			usermgr_uci_fini(DMMAP_SYSTEM_USER_FILE);
		}
	}
}

static void create_groups(void)
{
	struct uci_section *s = NULL;

	usermgr_uci_init(STD_UCI_PATH);

	usermgr_uci_foreach_section("users", "group", s) {
		char *group_name = NULL;
		const char *val = NULL;

		int len = strlen(GROUP_PREFIX);

		if (strncmp(s->e.name, GROUP_PREFIX, len) == 0)
			group_name = strdup(s->e.name + len);
		else
			group_name = strdup(s->e.name);

		val = usermgr_uci_get_value_by_section(s, "enabled");
		bool enable = dmuci_string_to_boolean(val);

		val = usermgr_uci_get_value_by_section(s, "deleted");
		bool deleted = dmuci_string_to_boolean(val);

		const char *old_name = usermgr_uci_get_value_by_section(s, "old_name");

		if (deleted) {
			del_group(group_name);
			usermgr_uci_delete_by_section(s, NULL, NULL);
			FREE(group_name);
			continue;
		}

		if (strlen(old_name) != 0) {
			del_group(old_name);
			usermgr_uci_set_value_by_section(s, "old_name", "");
		}

		if (enable) {
			add_group(group_name);
		} else {
			del_group(group_name);
		}

		FREE(group_name);
	}

	usermgr_uci_fini("users");
}

static void create_users(void)
{
	struct uci_section *s = NULL;

	usermgr_uci_init(STD_UCI_PATH);

	usermgr_uci_foreach_section("users", "user", s) {
		const char *val = NULL;
		char *user_name = strdup(s->e.name);

		val = usermgr_uci_get_value_by_section(s, "enabled");
		bool enable = dmuci_string_to_boolean(val);

		val = usermgr_uci_get_value_by_section(s, "deleted");
		bool deleted = dmuci_string_to_boolean(val);

		const char *old_name = usermgr_uci_get_value_by_section(s, "old_name");
		const char *member_groups = usermgr_uci_get_value_by_section(s, "member_groups");
		const char *shell = usermgr_uci_get_value_by_section(s, "shell");
		const char *enc_password = usermgr_uci_get_value_by_section(s, "encrypted_password");

		if (deleted) {
			del_user(user_name);
			usermgr_uci_delete_by_section(s, NULL, NULL);
			continue;
		}

		if (strlen(old_name) != 0) {
			change_username(old_name, user_name);
			usermgr_uci_set_value_by_section(s, "old_name", "");
		}

		if (!user_exist(user_name)) {
			add_user(user_name);
		}

		if (strlen(enc_password) != 0) {
			change_password(user_name, enc_password);
		}

		set_user_shell(user_name, shell);
		set_user_group(user_name, member_groups);

		if (enable)
			enable_user(user_name);
		else
			disable_user(user_name);

		FREE(user_name);
	}

	usermgr_uci_fini("users");
}

static void config_reload_cb(struct ubus_context *ctx __attribute__((unused)),
			struct ubus_event_handler *ev __attribute__((unused)),
			const char *type __attribute__((unused)),
			struct blob_attr *msg __attribute__((unused)))
{
	create_groups();
	create_users();
}

static int register_config_change(struct ubus_context *uctx)
{
	int ret;

	memset(&ev, 0, sizeof(struct ubus_event_handler));
	ev.cb = config_reload_cb;

	ret = ubus_register_event_handler(uctx, &ev, "usermngr.reload");
	if (ret)
		return -1;

	return 0;
}

int main(int argc, char **argv)
{
	char fname[128] = {0};
	snprintf(fname, sizeof(fname), "%s/%s", DMMAP_UCI_PATH, DMMAP_SYSTEM_USER_FILE);
	if (!file_exists(fname)) {
		// cppcheck-suppress cert-MSC24-C
		FILE *fp = fopen(fname, "w");
		if (fp) {
			fclose(fp);
		}
	}

	memset(&bbfdm_ctx, 0, sizeof(struct bbfdm_context));

	sync_system_user();
	sync_system_group();
	sync_static_role();

	create_groups();
	create_users();

	bbfdm_ubus_set_service_name(&bbfdm_ctx, "usermngr");
	bbfdm_ubus_set_log_level(3);
	bbfdm_ubus_load_data_model(tDynamicObj);

	if (bbfdm_ubus_regiter_init(&bbfdm_ctx))
		goto out;

	if (register_config_change(&bbfdm_ctx.ubus_ctx) != 0)
		goto out;

	uloop_run();

out:
	bbfdm_ubus_regiter_free(&bbfdm_ctx);
	return 0;
}
